/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */




/************************************************************************
 * sig.c: signal mapping functions for libthread.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include "threadP.h"




#if SIG_READY

/************************************************************************
 * macros.
************************************************************************/

#define is_invalid_sig(s) (((s)<=0)||((s)>=NSIG))





/************************************************************************
 * global variables.
************************************************************************/

static unsigned in_cs = 0;
static sigset_t omask;





#if defined(ENABLE_PREEMPTION)
/************************************************************************
 * void thr_preemption_handler(int signo)
 *
 * handle preemption of threads on the receipt of SIGVTALRM.
************************************************************************/

static void
thr_preemption_handler(int signo)
{
    if (thr_inlib > 0) {
        return;
    }

    assert(signo == SIGVTALRM);
    assert(thr_curr);

    if (thr_curr->flags.blocked) {
        printf("curr %d blocked??\n", thr_curr->id.tid);
        return;
    }

    ENQUEUE(thr_ready_q, thr_curr);
    thr_reschedule();

    return;
}





/************************************************************************
 * void thr_enable_preemption(void)
 *
 * install the preemption handler for libthread.
************************************************************************/

void thr_enable_preemption(void)
{
    struct sigaction act;
    struct itimerval it;


    act.sa_handler = (Sa_Handler) thr_preemption_handler;
    act.sa_flags   = 0;
#if defined(SA_INTERRUPT)
    act.sa_flags  |= SA_INTERRUPT;
#endif /* defined(SA_INTERRUPT) */
    sigfillset(&act.sa_mask);
    if (sigaction(SIGVTALRM, &act, 0) == -1) {
        perror("sigaction");
        abort();
    }

    it.it_value.tv_sec  = it.it_interval.tv_sec  = 0;
    it.it_value.tv_usec = it.it_interval.tv_usec = thr_quantum;
    if (setitimer(ITIMER_VIRTUAL, &it, 0) == -1) {
        perror("setitimer");
        abort();
    }

    return;
}
#endif /* defined(ENABLE_PREEMPTION) */





/************************************************************************
 * void thr_sig_handler(int signo)
 *
 * the common libthread signal handler for all signals. on invocation,
 * increment a counter for the current signal.
 *
 * signals need to be converted to messages during message reception.
************************************************************************/

static void
thr_sig_handler(int signo)
{
    assert(SIG_MAP(signo)->bound);

    SIG_MAP(signo)->count++;
    ITEM_TAB(SIG_MAP(signo)->tid)->sig.nsigs++;

    return;
}





/************************************************************************
 * void thr_init_sig_map(unsigned i)
 *
 * clear out the fields of the signal map structure for entry `i'.
************************************************************************/

void
thr_init_sig_map(unsigned i)
{
    thr_sig_map[i].bound = 0;
    thr_sig_map[i].tid   = 0;
    thr_sig_map[i].count = 0;

    return;
}





/************************************************************************
 * void thr_alloc_sig_map(void)
 *
 * allocate the signal map structure and clear out the fields.
************************************************************************/

void
thr_alloc_sig_map(void)
{
    unsigned i;

    thr_sig_map = (Sig_Map)
        malloc(sizeof(struct _Sig_Map)*thr_sig_map_size);
    if (!thr_sig_map) {
        perror("malloc");
        abort();
    }

    for (i = 0; i < thr_sig_map_size; i++) {
        thr_init_sig_map(i);
    }

    return;
}





/************************************************************************
 * void thr_sig_enter_cs(void)
 * void thr_sig_leave_cs(void)
 *
 * enter and leave a critical section which is not interruptible by
 * signals. all signals are blocked and unblocked (rspy).
************************************************************************/

void
thr_sig_enter_cs(void)
{
    sigset_t mask;

    assert(!in_cs);
    sigfillset(&mask);
    if (sigprocmask(SIG_BLOCK, &mask, &omask) == -1) {
        perror("sigprocmask(SIG_BLOC)");
        abort();
    }
    in_cs = 1;

    DO_TRACE_1(("thr_sig_enter_cs(): entering critical section"));
    return;
}



void
thr_sig_leave_cs(void)
{
    assert(in_cs);
    if (sigprocmask(SIG_SETMASK, &omask, 0) == -1) {
        perror("sigprocmask(SIG_SETMASK)");
        abort();
    }
    in_cs = 0;

    DO_TRACE_1(("thr_sig_leave_cs(): leaving critical section"));
    return;
}





/************************************************************************
 * void thr_unbind_sigs(void)
 *
 * the current thread is exiting. unbind all signals bound to this
 * thread.
************************************************************************/

void
thr_unbind_sigs(void)
{
    unsigned i;

    assert(thr_curr);
    DO_TRACE_1(("thr_unbind_sigs(): unbinding all signals for dying thread"));

    for (i = 0; i < thr_sig_map_size; i++) {
        if (i_own_sig(i)) {
            sig_unbind((int) ndx2sig(i));
        }
    }

    return;
}





/************************************************************************
 * unsigned thr_check_sig_msg(void)
 *
 * this function is called when a `select' is interrupted by a signal.
 * check all bound signals for occurrences. if the thread bound to such
 * a signal is blocked waiting for a signal message, then convert
 * signal to a message.
 *
 * return `1' is a signal was converted into a message, `0' otherwise.
 * a return of `0' indicates that the signal which interrupted the
 * `select' was either unbound in this process or was for a thread that
 * was not waiting for signal messages. the `select' should be retried.
************************************************************************/

unsigned
thr_check_sig_msg(void)
{
    unsigned i;
    unsigned ret;
    Thread   t;


    DO_TRACE_1(("thr_check_sig_msg(): checking for global signals"));
    thr_sig_enter_cs();

    ret = 0;
    for (i = 0; i < thr_sig_map_size; i++) {
        if (thr_sig_map[i].bound && (thr_sig_map[i].count > 0)) {
            t = ITEM_TAB(thr_sig_map[i].tid);
            if (t->flags.blocked && is_sig_tag(*(t->msg.tagp))) {
                *(t->msg.tidp)    = ndx2sig(i);
                *(t->msg.tagp)    = MSG_TAG_SIG;
                if (!(t->flags.polling)) {
                    *(t->msg.nitemsp)    = thr_sig_map[i].count;
                    t->sig.nsigs        -= thr_sig_map[i].count;
                    thr_sig_map[i].count = 0;
                }

                t->flags.blocked = 0;
                t->flags.polling = 1;
                ENQUEUE(thr_ready_q, t);

                ret = 1;
                break;
            }
        }
    }

    thr_sig_leave_cs();
    return ret;
}





/************************************************************************
 * int sig_bind(int signo)
 *
 * bind signal `signo' to the current thread. any `signo' signals that
 * the current process receives will appear as messages to the calling
 * thread (check `msg_recv()' for an explanation of what exactly these
 * messages look like).
 *
 * the existing disposition of the signal `signo' is saved. when the
 * signal is later unbound by the calling thread, the saved disposition
 * is restored.
 *
 * it is an error to remap bound signals or bind a signal that is
 * currently bound to other threads.
************************************************************************/

int
sig_bind(int signo)
{
    struct sigaction act;
    int              ret;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

    if (is_invalid_sig(signo)) {
        fprintf(stderr, "invalid signal number `%d'\n", signo);
        RETURN(THR_EBADSIG);
    }

    NOT_VTALRM(signo);

    if (SIG_MAP(signo)->bound) {
        fprintf(stderr, "signal `%d' already bound to thread `%u'\n",
            signo, SIG_MAP(signo)->tid);
        RETURN(THR_EBADSIG);
    }

    SIG_MAP(signo)->bound = 1;
    SIG_MAP(signo)->tid   = thr_curr->id.tid;
    SIG_MAP(signo)->count = 0;

#if defined(i386_unknown_linux2_0) || defined(ia64_unknown_linux2_4)
    // Don't want to cast this in Linux
    act.sa_handler = thr_sig_handler;
#else
    act.sa_handler = (Sa_Handler) thr_sig_handler;
#endif
    act.sa_flags   = 0;
#if defined(SA_INTERRUPT)
    act.sa_flags  |= SA_INTERRUPT;
#endif /* defined(SA_INTERRUPT) */
    sigfillset(&act.sa_mask);
    if (sigaction(signo, &act, &(SIG_MAP(signo)->oact)) == -1) {
        perror("sigaction");
        abort();
    }

    CHECK_RESCHEDULE();
    DO_TRACE_0(("sig_bind(%d): signal bound", signo));
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int sig_unbind(int signo)
 *
 * unbind the signal `signo' from the calling thread. any currently
 * queued signals will be discarded and the previous disposition of
 * the signal will be restored.
 *
 * it is an error to unbind non-bound signals or unbind signals that
 * are bound to other threads.
 *
 * when a thread dies, all signals bound to it are automatically
 * released (it is equivalent to the thread having done separate
 * unbinds on all signals that are bound to it).
************************************************************************/

int
sig_unbind(int signo)
{
    int ret;


    if (!thr_init_done) {
        fprintf(stderr, "*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (is_invalid_sig(signo)) {
        fprintf(stderr, "invalid signal `%d'\n", signo);
        RETURN(THR_EBADSIG);
    }

    NOT_VTALRM(signo);

    if (!SIG_MAP(signo)->bound) {
        fprintf(stderr, "signal `%d' is unbound\n", signo);
        RETURN(THR_EBADSIG);
    }

    if (SIG_MAP(signo)->tid != thr_curr->id.tid) {
        fprintf(stderr, "signal `%d' is bound to thread `%u'\n",
            signo, SIG_MAP(signo)->tid);
        RETURN(THR_EBADSIG);
    }

    if (sigaction(signo, &(SIG_MAP(signo)->oact), 0) == -1) {
        perror("sigaction");
        abort();
    }

    thr_init_sig_map((unsigned) sig2ndx(signo));

    CHECK_RESCHEDULE();
    DO_TRACE_0(("sig_unbind(%d): signal unbound", signo));
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}

#endif // SIG_READY


#if defined(THREADLIB_DEBUG)
void
Signal::WriteTo( ostream& s ) const
{
	// dump our base class data
	Bindable::WriteTo( s );

	// dump our data
	s << "signal = " << sig << endl;
	s << "count = " << count << endl;
}
#endif // defined(THREADLIB_DEBUG)


void
Signal::HandleMessage( void* buf, unsigned* sizep )
{
#if SIG_READY
#endif // SIG_READY
}





